<?php

namespace Modules\PLA\Controllers;

class Retirada {
    protected $connection;
    protected $environment;
    protected $newCodeNT;
    protected $Utils;
    protected $prefixo;
    protected $retiradaPorPrato;
    protected $controlaAlmoxarifado;
    protected $controlaLocal;
    protected $controlaLote;
    protected $visualizaSaldo;
    protected $precisao;
    protected $ValidacoesEstoqueService;
    
    public function __construct($connection, $environment, $newCodeNT, $UtilsIntegracao, $ValidacoesEstoqueService){
        $this->connection = $connection;
        $this->environment = $environment;
        $this->newCodeNT = $newCodeNT;
        $this->Utils = $UtilsIntegracao;
        $this->ValidacoesEstoqueService = $ValidacoesEstoqueService;
        
        $this->prefixo = "";
        $this->retiradaPorPrato = false;
        $this->controlaAlmoxarifado = false;
        $this->controlaLocal = false;
        $this->controlaLote = false;
        $this->visualizaSaldo = false;
        $this->precisao = 5;
    }
    
    //Inclusão de mensagens de erro / aviso
    private function incluiMensagem($msg, $erro=false){
        $this->Utils->incluiMensagem($this->prefixo . $msg, $erro);
    }
    
    //Consulta e retorna uma lista de serviços para a filial (todos habilitados para retiradas)
    private function consultaServicos($cdfilial){
        $query = "
            SELECT CDFILIAL, CDSERVFILI, NMSERVICO
              FROM SERVICO
             WHERE CDFILIAL = :CDFILIAL
               AND IDSERVATIV <> 'N'
             ORDER BY CDSERVFILI
        ";
        $params = [
            "CDFILIAL" => $cdfilial
        ];
        $prepared = $this->connection->prepare($query);
        $prepared->execute($params);
        return $prepared->fetchAll();
    }
    //Consulta dados da filial relacionados à retirada
    private function consultaFilial($cdfilial){
        $query = "
            SELECT IDCNTRRESC
              FROM FILIAL
             WHERE CDFILIAL = :CDFILIAL
        ";
        $prepared = $this->connection->prepare($query);
        $params = ["CDFILIAL" => $cdfilial];
        $prepared->execute($params);
        return $prepared->fetch();
    }
    //Consulta informações do serviço indicado
    private function consultaServicoparaRetirada($cdfilial, $cdservfili){
        $query = "
            SELECT CDFILIAL, CDSERVFILI, NMSERVICO, IDUTRETIPRAT
              FROM SERVICO
             WHERE CDFILIAL = :CDFILIAL
               AND CDSERVFILI = :CDSERVFILI
               AND IDSERVATIV <> 'N'
        ";
        $params = [
            "CDFILIAL" => $cdfilial,
            "CDSERVFILI" => $cdservfili
        ];
        $prepared = $this->connection->prepare($query);
        $prepared->execute($params);
        return $prepared->fetch();
    }
    //Consulta dados do cardápio indicado
    private function consultaCardapio($cdfilial, $cdservfili, $dtcardapio){
        $query = "
            SELECT CDFILIAL, CDSERVFILI, NRCARDAPIO
              FROM CARDADIA
             WHERE CDFILIAL = :CDFILIAL
               AND CDSERVFILI = :CDSERVFILI
               AND DTCARDAPIO = TO_DATE(:DTCARDAPIO, 'YYYYMMDD')
        ";
        $params = [
            "CDFILIAL" => $cdfilial,
            "CDSERVFILI" => $cdservfili,
            "DTCARDAPIO" => $dtcardapio
        ];
        $prepared = $this->connection->prepare($query);
        $prepared->execute($params);
        return $prepared->fetch();
    }
    //Insere cardápio no banco de dados
    private function insereCardapio($cdfilial, $cdservfili, $dtcardapio, $cdoperador){
        $nrcardapio = $this->newCodeNT->getCodeByTableName('CARDAPIO' . $cdfilial, $this->environment->getNrOrg());
        $query = "
            INSERT INTO CARDADIA
            (CDFILIAL, NRCARDAPIO, CDSERVFILI, CDOPERADOR,
             DTCARDAPIO,DTMESCARD,
             NRORG, NRORGULTATU, NRORGINCLUSAO,
             DTULTATU, DTINCLUSAO,
             CDOPERULTATU, CDOPERINCLUSAO)
            VALUES
            (:CDFILIAL, :NRCARDAPIO, :CDSERVFILI, :CDOPERADOR,
             TO_DATE(:DTCARDAPIO, 'YYYYMMDD'), TRUNC(TO_DATE(:DTCARDAPIO, 'YYYYMMDD'), 'MM'),
             :NRORG, :NRORG, :NRORG,
             SYSDATE, SYSDATE,
             :CDOPERADORLOGADO, :CDOPERADORLOGADO)
        ";
        $params = [
            "CDFILIAL" => $cdfilial,
            "NRCARDAPIO" => $nrcardapio,
            "CDSERVFILI" => $cdservfili,
            "CDOPERADOR" => $cdoperador,
            "DTCARDAPIO" => $dtcardapio,
            "NRORG" => $this->environment->getNrOrg(),
            "CDOPERADORLOGADO" => $this->environment->getCdOperador()
        ];
        $prepared = $this->connection->prepare($query);
        $prepared->execute($params);
        return $nrcardapio;
    }
    //Consulta o número de cardápio indicado, se não existir cria um
    private function consultaOuInsereNrCardapio($cdfilial, $cdservfili, $dtcardapio, $cdoperador){
        $cardapio = $this->consultaCardapio($cdfilial, $cdservfili, $dtcardapio);
        if(!$cardapio){
            return $this->insereCardapio($cdfilial, $cdservfili, $dtcardapio, $cdoperador);
        }
        else{
            return $cardapio['NRCARDAPIO'];
        }
    }
    //Consulta dados da retirada indicada
    private function consultaRetirada($cdfilial, $nrcardapio, $dtretirada){
        if(empty($nrcardapio)) return false;
        $query = "
            SELECT CDFILIAL, NRCARDAPIO, NRRETIRADA, CDFILANCESTQ, NRLANCESTQ
              FROM RETIRADA
             WHERE CDFILIAL = :CDFILIAL
               AND NRCARDAPIO = :NRCARDAPIO
               AND DTRETIRADA = TO_DATE(:DTRETIRADA, 'YYYYMMDD')
        ";
        $params = [
            "CDFILIAL" => $cdfilial,
            "NRCARDAPIO" => $nrcardapio,
            "DTRETIRADA" => $dtretirada
        ];
        $prepared = $this->connection->prepare($query);
        $prepared->execute($params);
        return $prepared->fetch();
    }
    //Insere novo lançamento em estoque no banco de dados
    private function insereLancamentoEstoque($cdfilial, $dtlancestq, $cdoperador, $dslancestq){
            $nrlancestq = $this->newCodeNT->getCodeByTableName('LANCTOEST'.$cdfilial, $this->environment->getNrOrg());
            $query = "
                INSERT INTO LANCTOESTOQ
                (CDFILIAL, CDFILIMOVI, NRLANCESTQ,
                 DTLANCESTQ, DSLANCESTQ, IDTPLANCTO,
                 IDALTCUSMED, IDSTATUSCTBEST, CDOPERADOR, 
                 NRORG, NRORGULTATU, NRORGINCLUSAO,
                 DTULTATU, DTINCLUSAO,
                 CDOPERULTATU, CDOPERINCLUSAO)
                VALUES
                (:CDFILIAL, :CDFILIAL, :NRLANCESTQ,
                 TO_DATE(:DTLANCESTQ, 'YYYYMMDD'), :DSLANCESTQ, :IDTPLANCTO,
                 :IDALTCUSMED, :IDSTATUSCTBEST, :CDOPERADOR,
                 :NRORG, :NRORG, :NRORG,
                 SYSDATE, SYSDATE,
                 :CDOPERADORLOGADO, :CDOPERADORLOGADO)
            ";
            $params = [
                "CDFILIAL" => $cdfilial,
                "NRLANCESTQ" => $nrlancestq,
                "DTLANCESTQ" => $dtlancestq,
                "DSLANCESTQ" => $dslancestq,
                "IDTPLANCTO" => '3',
                "IDALTCUSMED" => 'N',
                "IDSTATUSCTBEST" => 'P',
                "CDOPERADOR" => $cdoperador,
                "NRORG" => $this->environment->getNrOrg(),
                "CDOPERADORLOGADO" => $this->environment->getCdOperador()
            ];
            $prepared = $this->connection->prepare($query);
            $prepared->execute($params);
            return $nrlancestq;
        }
    //Apaga lançamento de estoque
    private function excluiLancamentoEstoque($cdfilial, $nrlancestq){
        $query = "
            DELETE FROM LANCTOESTOQ
             WHERE CDFILIAL = :CDFILIAL
               AND NRLANCESTQ = :NRLANCESTQ
        ";
        $params = [
            "CDFILIAL" => $cdfilial,
            "NRLANCESTQ" => $nrlancestq
        ];
        $prepared = $this->connection->prepare($query);
        $prepared->execute($params);    
    }
    //Insere nova retirada no banco de dados
    private function insereRetirada($cdfilial, $nrcardapio, $dtretirada, $cdoperador, $nrlancestq){
        $nrretirada = $this->newCodeNT->getCodeByTableName('RETIRADA' . $cdfilial, $this->environment->getNrOrg());
        $retiradaPorPrato = ($this->retiradaPorPrato) ? 'S' : 'N';
        $query = "
            INSERT INTO RETIRADA
            (CDFILIAL, NRCARDAPIO, NRRETIRADA, NRLANCESTQ,
             DTRETIRADA, NRREQUEXTRRE, CDFILANCESTQ, IDRETIPRAT,
             NRORG, NRORGULTATU, NRORGINCLUSAO,
             DTULTATU, DTINCLUSAO,
             CDOPERULTATU, CDOPERINCLUSAO)
            VALUES 
            (:CDFILIAL, :NRCARDAPIO, :NRRETIRADA, :NRLANCESTQ,
             TO_DATE(:DTRETIRADA, 'YYYYMMDD'), :NRREQUEXTRRE, :CDFILIAL, :IDRETIPRAT,
             :NRORG, :NRORG, :NRORG,
             SYSDATE, SYSDATE,
             :CDOPERADORLOGADO, :CDOPERADORLOGADO)
        ";
        $params = [
            "CDFILIAL" => $cdfilial,
            "NRCARDAPIO" => $nrcardapio,
            "NRRETIRADA" => $nrretirada,
            "NRLANCESTQ" => $nrlancestq,
            "DTRETIRADA" => $dtretirada,
            "NRREQUEXTRRE" => "0000000000",
            "IDRETIPRAT" => $retiradaPorPrato,
            "NRORG" => $this->environment->getNrOrg(),
            "CDOPERADORLOGADO" => $this->environment->getCdOperador()
        ];
        $prepared = $this->connection->prepare($query);
        $prepared->execute($params);    
    }
    //Apaga a retirada do banco
    private function excluiRetirada($cdfilial, $nrcardapio, $nrretirada){
        $query = "
            DELETE FROM RETIRADA
             WHERE CDFILIAL = :CDFILIAL
               AND NRCARDAPIO = :NRCARDAPIO
               AND NRRETIRADA = :NRRETIRADA
        ";
        $params = [
            "CDFILIAL" => $cdfilial,
            "NRCARDAPIO" => $nrcardapio,
            "NRRETIRADA" => $nrretirada
        ];
        $prepared = $this->connection->prepare($query);
        $prepared->execute($params);    
    }
    //Consulta a retirada indicada, inclui caso não exista 
    private function consultaOuInsereRetirada($cdfilial, $cdservfili, $dtretirada, $cdoperador, $dslancestq){
        $nrcardapio = $this->consultaOuInsereNrCardapio($cdfilial, $cdservfili, $dtretirada, $cdoperador);
        $retirada = $this->consultaRetirada($cdfilial, $nrcardapio, $dtretirada);
        if(!$retirada){
            $nrlancestq = $this->insereLancamentoEstoque($cdfilial, $dtretirada, $cdoperador, $dslancestq);
            $this->insereRetirada($cdfilial, $nrcardapio, $dtretirada, $cdoperador, $nrlancestq);
            $retirada = $this->consultaRetirada($cdfilial, $nrcardapio, $dtretirada);
        }
        return $retirada;
    }
    //Consulta parêmetros da paramgeral relacionados à retirada
    private function consultaParamgeral(){
        $query = "SELECT IDUTILDTPROC FROM PARAMGERAL";
        $prepared = $this->connection->prepare($query);
        $prepared->execute();
        return $prepared->fetch();
    }
    //Consulta parêmetros da paramestoque relacionados à retirada
    private function consultaParamestoque(){
        $query = "
            SELECT PE.IDLANCDTMEN,
                   PE.IDLANCDTMAI,
                   NVL(PE.NRDIASSAID, 0) AS NRDIASSAID,
                   NVL(PE.NRDIASSAIDPD, 0) AS NRDIASSAIDPD
              FROM PARAMESTOQUE PE
        ";
        $prepared = $this->connection->prepare($query);
        $prepared->execute();
        return $prepared->fetch();
    }
    //Consulta parêmetros da paramfilial relacionados à retirada
    private function consultaParamfilial($cdfilial){
        $query = "
            SELECT PF.IDLANCDTMENFIL,
                   PF.IDLANCDTMAIFIL,
                   PF.IDUTILALMOX,
                   PF.IDUTILLCESTQ,
                   PF.IDUTILLOTE,
                   TO_CHAR(PF.DTPROCESSA, 'YYYYMMDD') AS DTPROCESSA,
                   NVL(PF.NRDIASSAIDFIL, 0) AS NRDIASSAIDFIL,
                   NVL(PF.NRDIASSAIDPDFIL, 0) AS NRDIASSAIDPDFIL,
                   PF.IDPERMVISPOSEST
              FROM PARAMFILIAL PF
             WHERE PF.CDFILIAL = :CDFILIAL
        ";
        $prepared = $this->connection->prepare($query);
        $prepared->execute(['CDFILIAL' => $cdfilial]);
        return $prepared->fetch();
    }
    //Consulta parâmetros da paramplaneja relacionados à retirada
    private function consultaParamPlaneja($cdfilial){
        $query = "
            SELECT IDUTREQUPRAT, CDTPRETIPADR
              FROM PARAMPLANEJA
             WHERE CDFILIAL = :CDFILIAL
        ";
        $prepared = $this->connection->prepare($query);
        $prepared->execute(['CDFILIAL' => $cdfilial]);
        return $prepared->fetch();    
    }
    //Consulta informações do produto indicado juntamente com dados da prodfili
    private function consultaProduto($cdproduto, $cdfilial){
        $query = "
        SELECT P.CDPRODUTO, P.CDPRODESTO, NVL(P.IDPRODATIVPR, 'S') AS IDPRODATIVPR, PF.CDFILIAL,
               NVL(PF.IDPRODATIV, 'A') AS IDPRODATIV, NVL(PF.IDCNTRESTOQ, IDCNTRESTOQ) AS IDCNTRESTOQ
          FROM PRODUTO P
               LEFT JOIN PRODFILI PF ON PF.CDPRODUTO = P.CDPRODUTO 
                                    AND PF.CDFILIAL = :CDFILIAL
         WHERE P.CDPRODUTO = :CDPRODUTO
        ";
        $prepared = $this->connection->prepare($query);
        $params = [
            'CDPRODUTO' => $cdproduto,
            'CDFILIAL' => $cdfilial
        ];
        $prepared->execute($params);
        return $prepared->fetch();
    }
    //Consulta o operador indicado no banco de dados
    private function consultaOperador($cdOperador){
        $query = "
            SELECT CDOPERADOR
              FROM OPERADOR
             WHERE CDOPERADOR = :CDOPERADOR
        ";
        $prepared = $this->connection->prepare($query);
        $prepared->execute(['CDOPERADOR' => $cdOperador]);
        return $prepared->fetch();
    }
    //Consulta o tipo de prato indicado no banco de dados
    private function consultaTipoPrato($cdtipoprat){
        $query = "
            SELECT CDTIPOPRAT
              FROM TIPOPRAT
             WHERE CDTIPOPRAT = :CDTIPOPRAT
        ";
        $prepared = $this->connection->prepare($query);
        $prepared->execute(['CDTIPOPRAT' => $cdtipoprat]);
        return $prepared->fetch();
    }
    //Consulta o prato indicado no banco de dados
    private function consultaPrato($cdfilauxprat, $cdprato){
        $query = "
            SELECT CDFILAUXPRAT, CDPRATO
              FROM PRATO
             WHERE CDFILAUXPRAT = :CDFILAUXPRAT
               AND CDPRATO = :CDPRATO
               AND CDNVPRATO = (SELECT MAX(CDNVPRATO) FROM PRATO)
        ";
        $prepared = $this->connection->prepare($query);
        $params = [
            'CDFILAUXPRAT' => $cdfilauxprat,
            'CDPRATO' => $cdprato
        ];
        $prepared->execute($params);
        return $prepared->fetch();
    }
    //Consulta item específico da retirada existente
    private function consultaItemReti($cdfilial, $nrcardapio, $nrretirada, $cdproduto, $nrlancestq, $cdtipoprat, 
                                      $cdfilauxprat, $cdprato, $cdalmoxarife, $cdlocalestoq, $nrloteestq, $nrsublote, $cdfilancestq){
        $retiradaPorPrato = ($this->retiradaPorPrato) ? 'S' : 'N';
        $query = "
            SELECT I.CDFILIAL, I.NRCARDAPIO, I.NRRETIRADA, I.NRSEQUITRETI, I.CDPRODUTO,
                   I.QTPREVISRETI, I.QTEXTRARETI, I.QTDEVOLVRETI, I.QTRETIRADA, I.NRSEQUITEM,
                   I.QTSALDORETIRADA, IL.QTLANCTOEST, IL.QTTOTLANCTO, IL.VRUNILANCTO, IL.VRTOTLANCTO,
                   IL.IDEVOLESTOQ
              FROM ITEMRETI I
                   LEFT JOIN ITLANCTOEST IL ON IL.CDFILIAL = I.CDFIITLANEST AND
                                               IL.NRLANCESTQ = I.NRLANCESTQ AND
                                               IL.NRSEQUITEM = I.NRSEQUITEM
             WHERE I.CDFILIAL = :CDFILIAL
               AND I.NRCARDAPIO = :NRCARDAPIO
               AND I.NRRETIRADA = :NRRETIRADA
               AND I.CDPRODUTO = :CDPRODUTO
               AND I.CDFIITLANEST = :CDFIITLANEST
               AND I.NRLANCESTQ = :NRLANCESTQ
               AND ((:RETIRADAPORPRATO = 'N') OR (I.CDTIPOPRAT = :CDTIPOPRAT))
               AND ((:RETIRADAPORPRATO = 'N') OR (I.CDFILAUXPRAT = :CDFILAUXPRAT))
               AND ((:RETIRADAPORPRATO = 'N') OR (I.CDPRATO = :CDPRATO))
               AND IL.CDALMOXARIFE = :CDALMOXARIFE
               AND IL.CDLOCALESTOQ = :CDLOCALESTOQ
               AND IL.NRLOTEESTQ = :NRLOTEESTQ
               AND IL.NRSUBLOTE = :NRSUBLOTE
        ";
        $prepared = $this->connection->prepare($query);
        $params = [
            'CDFILIAL' => $cdfilial,
            'NRCARDAPIO' => $nrcardapio,
            'NRRETIRADA' => $nrretirada,
            'CDPRODUTO' => $cdproduto,
            'CDFIITLANEST' => $cdfilancestq,
            'NRLANCESTQ' => $nrlancestq,
            'RETIRADAPORPRATO' => $retiradaPorPrato,
            'CDTIPOPRAT' => $cdtipoprat,
            'CDFILAUXPRAT' => $cdfilauxprat,
            'CDPRATO' => $cdprato,
            'CDALMOXARIFE' => $cdalmoxarife,
            'CDLOCALESTOQ' => $cdlocalestoq,
            'NRLOTEESTQ' => $nrloteestq,
            'NRSUBLOTE' => $nrsublote
        ];
        $prepared->execute($params);
        return $prepared->fetch();    
    }
    //Consulta os itens de retirada. É possível filtrar apenas os que apresentaram erro de estoque
    private function consultaItensReti($cdfilial, $nrcardapio, $nrretirada, $comErroEstoque=false){
        $erroEmEstoque = ($comErroEstoque) ? 'N' : null;
        $query = "
            SELECT I.NRSEQUITRETI 
              FROM ITEMRETI I
                   LEFT JOIN ITLANCTOEST IL ON IL.CDFILIAL = I.CDFIITLANEST AND
                                               IL.NRLANCESTQ = I.NRLANCESTQ AND
                                               IL.NRSEQUITEM = I.NRSEQUITEM
             WHERE I.CDFILIAL = :CDFILIAL
               AND I.NRCARDAPIO = :NRCARDAPIO
               AND I.NRRETIRADA = :NRRETIRADA
               AND ((IL.IDEVOLESTOQ = :IDEVOLESTOQ) OR (:IDEVOLESTOQ IS NULL))
        ";
        $prepared = $this->connection->prepare($query);
        $params = [
            'CDFILIAL' => $cdfilial,
            'NRCARDAPIO' => $nrcardapio,
            'NRRETIRADA' => $nrretirada,
            'IDEVOLESTOQ' => $erroEmEstoque
        ];
        $prepared->execute($params);
        return $prepared->fetchAll();
    }
    //Consulta a próxima numeração de item da retirada
    private function consultaProxNrItem($cdfilial, $nrcardapio, $nrretirada){
        $query = "
            SELECT MAX(NRSEQUITRETI) AS NRSEQUITRETI
              FROM ITEMRETI
             WHERE CDFILIAL = :CDFILIAL
               AND NRCARDAPIO = :NRCARDAPIO
               AND NRRETIRADA = :NRRETIRADA
        ";
        $prepared = $this->connection->prepare($query);
        $params = [
            'CDFILIAL' => $cdfilial,
            'NRCARDAPIO' => $nrcardapio,
            'NRRETIRADA' => $nrretirada
        ];
        $prepared->execute($params);
        $numero = $prepared->fetch();     
        $numero = $numero['NRSEQUITRETI'];
        if(empty($numero)){
            $numero = 0;
        }
        else{
            try{
                $numero = intval($numero) + 1;
            }
            catch(\Exception $e){
                $numero = 0;
            }
        }
        return $numero;
    }
    //Retorna data da última posição do produto (considera como limite um dia após a data da retirada)
    private function retornaUltimaDataPosicaoEstoque($cdfilial, $dtretirada, $cdproduto, $cdalmoxarife, $cdlocalestoq, $nrloteestq, $nrsublote){
        $query = "
            SELECT TO_CHAR(MAX(DTPOSIESTQ), 'YYYYMMDD') AS DTPOSIESTQ
              FROM POSESTDIA
             WHERE CDFILIAL = :CDFILIAL
               AND CDPRODUTO = :CDPRODUTO
               AND CDALMOXARIFE = :CDALMOXARIFE
               AND CDLOCALESTOQ = :CDLOCALESTOQ
               AND NRLOTEESTQ = :NRLOTEESTQ
               AND NRSUBLOTE = :NRSUBLOTE
               AND DTPOSIESTQ <= TO_DATE(:DTPOSIESTQ, 'YYYYMMDD') + 1
        ";
        $prepared = $this->connection->prepare($query);
        $params = [
            'CDFILIAL' => $cdfilial,
            'CDPRODUTO' => $cdproduto,
            'CDALMOXARIFE' => $cdalmoxarife,
            'CDLOCALESTOQ' => $cdlocalestoq,
            'NRLOTEESTQ' => $nrloteestq,
            'NRSUBLOTE' => $nrsublote,
            'DTPOSIESTQ' => $dtretirada
        ];
        $prepared->execute($params);
        $posestdia = $prepared->fetch();
        return $posestdia['DTPOSIESTQ'];
        
    }
    //Consulta e retorna o saldo na retirada do produto no inicio do dia posterior a retirada (ou seja considerando todos os lançamentos já feitos no dia)
    private function consultaSaldoDia($cdfilial, $dtretirada, $cdproduto, $cdalmoxarife, $cdlocalestoq, $nrloteestq, $nrsublote){
        $dtultpos = $this->retornaUltimaDataPosicaoEstoque($cdfilial, $dtretirada, $cdproduto, $cdalmoxarife, $cdlocalestoq, $nrloteestq, $nrsublote);
        $query = "
            SELECT QTESTOQDIA
              FROM POSESTDIA
             WHERE CDFILIAL = :CDFILIAL
               AND CDPRODUTO = :CDPRODUTO
               AND CDALMOXARIFE = :CDALMOXARIFE
               AND CDLOCALESTOQ = :CDLOCALESTOQ
               AND NRLOTEESTQ = :NRLOTEESTQ
               AND NRSUBLOTE = :NRSUBLOTE
               AND DTPOSIESTQ = TO_DATE(:DTPOSIESTQ, 'YYYYMMDD')
        ";
        $prepared = $this->connection->prepare($query);
        $params = [
            'CDFILIAL' => $cdfilial,
            'CDPRODUTO' => $cdproduto,
            'CDALMOXARIFE' => $cdalmoxarife,
            'CDLOCALESTOQ' => $cdlocalestoq,
            'NRLOTEESTQ' => $nrloteestq,
            'NRSUBLOTE' => $nrsublote,
            'DTPOSIESTQ' => $dtultpos
        ];
        $prepared->execute($params);
        $posestdia = $prepared->fetch();
        if(!$posestdia) return 0;
        return $posestdia['QTESTOQDIA'];
        
    }
    //Obtém custo de estoque do produto
    private function consultaCustoEstoque($cdfilial, $cdproduto, $dtlancto, $cdalmoxarife, $cdlocalestoq, $nrloteestq, $nrsublote){
        $query = "
            SELECT CUSTO_PRODUTO_LCEST(
            :CDFILIAL, :CDPRODUTO, TO_NUMBER(NVL(CDTIPCUSTEST, '3')),
            TO_DATE(:DTPOSICAO, 'YYYYMMDD'), SYSDATE, :CDALMOXARIFE, :CDLOCALESTOQ,
            :NRLOTEESTQ, :NRSUBLOTE) AS CUSTO
              FROM PARAMFILIAL
             WHERE CDFILIAL = :CDFILIAL
        ";
        $prepared = $this->connection->prepare($query);
        $params = [
            'CDFILIAL' => $cdfilial,
            'CDPRODUTO' => $cdproduto,
            'DTPOSICAO' => $dtlancto,
            'CDALMOXARIFE' => $cdalmoxarife,
            'CDLOCALESTOQ' => $cdlocalestoq,
            'NRLOTEESTQ' => $nrloteestq,
            'NRSUBLOTE' => $nrsublote
        ];
        $prepared->execute($params);
        $custo = $prepared->fetch();
        if(!$custo) return 0.0;
        return $custo['CUSTO'];
    }
    //Consulta almoxarifado no banco de dados
    private function consultaAlmoxarifado($cdfilial, $cdalmoxarife){
        $query = "
        SELECT CDALMOXARIFE 
          FROM ALMOXARIFADO
         WHERE CDFILIAL = :CDFILIAL
           AND CDALMOXARIFE = :CDALMOXARIFE
        ";
        $prepared = $this->connection->prepare($query);
        $params = [
            "CDFILIAL" => $cdfilial,
            "CDALMOXARIFE" => $cdalmoxarife
        ];
        $prepared->execute($params);
        return $prepared->fetch();
    }
    //Consulta local de estoque no banco de dados
    private function consultaLocalEstoque($cdfilial, $cdalmoxarife, $cdlocalestoq){
        $query = "
            SELECT CDLOCALESTOQ
              FROM LOCALESTOQUE
             WHERE CDFILIAL = :CDFILIAL
               AND CDALMOXARIFE = :CDALMOXARIFE
               AND CDLOCALESTOQ = :CDLOCALESTOQ
        ";
        $prepared = $this->connection->prepare($query);
        $params = [
            "CDFILIAL" => $cdfilial,
            "CDALMOXARIFE" => $cdalmoxarife,
            "CDLOCALESTOQ" => $cdlocalestoq
        ];
        $prepared->execute($params);
        return $prepared->fetch();    
    }
    //Consulta almoxarifado no banco de dados
    private function consultaLote($cdfilial, $dtretirada, $cdproduto, $nrloteestq, $nrsublote){
        $query = "
        SELECT NRLOTEESTQ
          FROM LOTE
         WHERE CDFILIAL = :CDFILIAL
           AND NRLOTEESTQ = :NRLOTEESTQ
           AND NRSUBLOTE = :NRSUBLOTE
           AND CDPRODUTO = :CDPRODUTO
           AND TO_DATE(:DTRETIRADA, 'YYYYMMDD') <= DTVALILOTE
        ";
        $prepared = $this->connection->prepare($query);
        $params = [
            "CDFILIAL" => $cdfilial,
            "NRLOTEESTQ" => $nrloteestq,
            "NRSUBLOTE" => $nrsublote,
            "CDPRODUTO" => $cdproduto,
            "DTRETIRADA" => $dtretirada
        ];
        $prepared->execute($params);
        return $prepared->fetch();
    }
    //Verifica se o parâmetro de filial é valido
    private function validaFilial($cdfilial){
        if(!$cdfilial){
            $this->incluiMensagem("Unidade não informada", true);
            return false;
        }
        $filial = $this->consultaFilial($cdfilial);
        if(!$filial){
            $this->incluiMensagem("Unidade não cadastrada", true);
            return false;
        }
        else if ($filial['IDCNTRRESC'] == 'S'){
            $this->incluiMensagem("Unidade inativa", true);
            return false;
        }
        return true;
    }
    //Verifica se o parâmetro de serviço é valido
    private function validaServico($cdfilial, $cdservfili){
        if(empty($cdservfili)){
            $this->incluiMensagem("Serviço não informado", true);
            return false;
        }
        $servico = $this->consultaServicoparaRetirada($cdfilial, $cdservfili);
        if(!$servico){
            $this->incluiMensagem("Serviço informado não cadastrado", true);
            return false;
        }
        if($servico['IDUTRETIPRAT'] == 'S'){
            $this->retiradaPorPrato = true;
        }
        return true;
    }
    //Verifica se o parâmetro de data é valido e se está no período permitido
    private function validaPeriodo($cdfilial, $dtretirada, $paramfilial){
        if(empty($dtretirada)){
            $this->incluiMensagem('Data não informada', true);
            return false;
        }
        $dtretirada = strtotime($dtretirada);
        if(!$dtretirada){
            $this->incluiMensagem('Data informada não é um valor válido', true);
            return false;
        }
        $paramgeral = $this->consultaParamgeral();
        $paramestoque = $this->consultaParamestoque();
        
        $idutildtproc = ($paramgeral) ? $paramgeral['IDUTILDTPROC'] : 'N';
        $dtprocessa = strtotime(date('Ymd'));

        $idlancdtmen = 'N';
        $idlancdtmai = 'N';
        $nrdiassaid = 0;
        $nrdiassaidpd = 0;
        if($paramestoque){
            $idlancdtmen = $paramestoque['IDLANCDTMEN'];
            $idlancdtmai = $paramestoque['IDLANCDTMAI'];
            $nrdiassaid = $paramestoque['NRDIASSAID'];
            $nrdiassaidpd = $paramestoque['NRDIASSAIDPD'];
        }
        
        $idlancdtmenfil = 'N';
        $nrdiassaidfil = 0;
        $idlancdtmaifil = 'N';
        $nrdiassaidpdfil = 0;
        
        if($paramfilial){
            if($idutildtproc = 'S' && !empty($paramfilial['DTPROCESSA'])){
                $dtprocessa = strtotime($paramfilial['DTPROCESSA']);
            }
            $idlancdtmenfil = $paramfilial['IDLANCDTMENFIL'];
            $nrdiassaidfil = $paramfilial['NRDIASSAIDFIL'];
            $idlancdtmaifil = $paramfilial['IDLANCDTMAIFIL'];
            $nrdiassaidpdfil = $paramfilial['NRDIASSAIDPDFIL'];
        }
        
        $diasAnterior = 0;
        if($idlancdtmenfil = 'S'){
            $diasAnterior = $nrdiassaidfil;
        } else if ($idlancdtmen = 'S'){
            $diasAnterior = $nrdiassaid;
        }
        
        $diasPosterior = 0;
        
        if($idlancdtmaifil = 'S'){
            $diasPosterior = $nrdiassaidpdfil;
        } else if ($idlancdtmai = 'S'){
            $diasPosterior = $nrdiassaidpd;
        }
        
        $dataAnterior = strtotime("- $diasAnterior day", $dtprocessa);
        $dataPosterior = strtotime("+ $diasPosterior day", $dtprocessa);
        
        if (($dtretirada < $dataAnterior) || ($dtretirada > $dataPosterior)){
            $this->incluiMensagem("Data informada fora do período permitido", true);
            return false;
        }
        
        return true;
    }
    //Verifica parametrizações de planejamento para a retirada. Atribui o tipo de retirada padrão.
    private function validaParamPlaneja($cdfilial, &$cdtpretipadr){
        $cdtpretipadr = null;
        $paramplaneja = $this->consultaParamPlaneja($cdfilial);
        if(!$paramplaneja){
            $this->incluiMensagem("Tipo de retirada padrão não cadastrado em parâmetros de planejamento", true);
            return true;
        } else{
            $idObrigaRetiradaPrato = ($paramplaneja['IDUTREQUPRAT'] == 'S') ? true : false;
            if($idObrigaRetiradaPrato && !$this->retiradaPorPrato){
                $this->incluiMensagem("Unidade obriga retirada por prato porém o serviço não está obrigando a retirada desta forma", true);
                return false;
            }
            $cdtpretipadr = $paramplaneja['CDTPRETIPADR'];
            if(empty($cdtpretipadr)){
                $this->incluiMensagem("Tipo de retirada padrão não cadastrado em parâmetros de planejamento", true);
            }
            return true;
        }
    }
    //Verifica se a lista de produtos informada é válida
    private function validaListaProdutos($produtos){
        if(!is_array($produtos)){
            $this->incluiMensagem("Produtos não informados");
            return false;
        }
        if(count($produtos) == 0){
            $this->incluiMensagem("Produtos não informados");
            return false;
        }
        return true;
    }
    //Verifica se o operador informado é um operador válido do sistema
    private function validaOperador($cdoperador){
        if(empty($cdoperador)){
            $this->incluiMensagem("Operador não informado", true);
            return false;
        }
        $operador = $this->consultaOperador($cdoperador);
        if(!$operador){
            $this->incluiMensagem("Operador informado não cadastrado", true);
            return false;
        }
        return true;
    }
    //Verifica se a descrição do lançamento de estoque é válida
    private function validaDescricaoEstoque($dslancestq){
        if(!$dslancestq){
            $this->incluiMensagem("Descrição do lançamento de estoque não informada", true);
            return false;
        }
        return true;    
    }
    /*Verifica se o produto informado é válido, contém todas as informações para realização da retirada
    Normaliza valores para realização da retirada do produto*/
    private function validaERetornaProduto($produto, $indice, $cdfilial, $nrcardapio, $nrretirada, $dtretirada, $nrlancestq, $cdtpretipadr, $cdfilancestq){
        $produtoValido = true;
        
        $cdproduto = $this->Utils->valor($produto, 'CDPRODUTO');
        $cdbarproduto = $this->Utils->valor($produto, 'CDBARPRODUTO');
        $qtPrevista = 0;
        $qtExtra = $this->Utils->valor($produto, 'QTEXTRARETI', 0.0);
        if($qtExtra == null)
            $qtExtra = 0;
        $qtExtra = round($qtExtra, $this->precisao);
        $qtDevolvida = $this->Utils->valor($produto, 'QTDEVOLVRETI', 0.0);
        if($qtDevolvida == null)
            $qtDevolvida = 0;
        $qtDevolvida = round($qtDevolvida, $this->precisao);
        $cdfilauxprat = $this->Utils->valor($produto, 'CDFILAUXPRAT');
        $cdprato = $this->Utils->valor($produto, 'CDPRATO');
        $cdtipoprat = $this->Utils->valor($produto, 'CDTIPOPRAT');
        
        $cdalmoxarife = $this->Utils->valor($produto, 'CDALMOXARIFE', ' ');
        $cdalmoxarife = (empty($cdalmoxarife)) ? ' ' : $cdalmoxarife;
        $cdlocalestoq = $this->Utils->valor($produto, 'CDLOCALESTOQ', ' ');
        $cdlocalestoq = (empty($cdlocalestoq)) ? ' ' : $cdlocalestoq;
        $nrloteestq = $this->Utils->valor($produto, 'NRLOTEESTQ', ' ');
        $nrloteestq = (empty($nrloteestq)) ? ' ' : $nrloteestq;
        $nrsublote = $this->Utils->valor($produto, 'NRSUBLOTE', ' ');
        $nrsublote = (empty($nrsublote)) ? ' ' : $nrsublote;
        
        $sobrepor = $this->Utils->valor($produto, 'sobrepor', false);
        
        $nrsequitreti = null;
        $nrsequitem = null;
        $qtPrevistaJaLancada = 0;
        $qtExtraJaLancada = 0;
        $qtDevolvidaJaLancada = 0;
        $qtRetiradaJaLancada = 0;
        $qtSaldoRetiradaJaLancado = 0;
        $qtlanctoest = 0;
        $qttotlancto = 0;
        $vrunilancto = 0;
        $vrtotlancto = 0;
        $idevolestoq = 'N';
        
        $produtoNovo = true;
        //Se item já existe na base, registra os dados antigos (que serão substituídos)
        $itemreti = $this->consultaItemReti($cdfilial, $nrcardapio, $nrretirada, $cdproduto, $nrlancestq, 
                                            $cdtipoprat, $cdfilauxprat, $cdprato,
                                            $cdalmoxarife, $cdlocalestoq, $nrloteestq, $nrsublote, $cdfilancestq);
        
        if($itemreti){
            $produtoNovo = false;
            $nrsequitreti = $itemreti['NRSEQUITRETI'];
            $nrsequitem = $itemreti['NRSEQUITEM'];
            $qtPrevistaJaLancada = ($itemreti['QTPREVISRETI'] == null) ? 0 : $itemreti['QTPREVISRETI'];
            $qtPrevista = $qtPrevistaJaLancada;
            $qtExtraJaLancada = ($itemreti['QTEXTRARETI'] == null) ? 0 : $itemreti['QTEXTRARETI'];
            $qtDevolvidaJaLancada = ($itemreti['QTDEVOLVRETI'] == null) ? 0 : $itemreti['QTDEVOLVRETI'];
            $qtRetiradaJaLancada = ($itemreti['QTRETIRADA'] == null) ? 0 : $itemreti['QTRETIRADA'];
            $qtSaldoRetiradaJaLancado = $itemreti['QTSALDORETIRADA'];
            
            $qtlanctoest = $itemreti['QTLANCTOEST'];
            $qttotlancto = $itemreti['QTTOTLANCTO'];
            $vrunilancto = $itemreti['VRUNILANCTO'];
            $vrtotlancto = $itemreti['VRTOTLANCTO'];
            $idevolestoq = $itemreti['IDEVOLESTOQ'];
        }
        
        $CUSTOMCLOB1 = $this->Utils->valor($produto, 'CUSTOMCLOB1');
        $CUSTOMCLOB2 = $this->Utils->valor($produto, 'CUSTOMCLOB2');
        //Rótulo do produto nas mensagens de erro
        $identificacao = "$cdproduto($indice)";
        if(empty($cdproduto)){
            $identificacao = "($indice)";
            $this->incluiMensagem("Produto $identificacao não possui código");
            $produtoValido = false;
        }
        
        if($qtExtra <= 0 && $qtDevolvida <= 0){
            $this->incluiMensagem("Produto $identificacao com quantidade extra ou devolvida não informada");
            $produtoValido = false;
        }
        
         //Valida tipo de prato e prato se o serviço obriga retirada por prato
        if($this->retiradaPorPrato){
            if(empty($cdtipoprat)){
                $this->incluiMensagem("Produto $identificacao - Tipo de prato não informado");
                $produtoValido = false;
            }else{
                $tipoprat = $this->consultaTipoPrato($cdtipoprat);
                if(!$tipoprat){
                    $this->incluiMensagem("Produto $identificacao - Tipo de prato informado não cadastrado");
                    $produtoValido = false;
                }
            }
            
            if(empty($cdfilauxprat) || empty($cdprato)){
                $this->incluiMensagem("Produto $identificacao - Prato não informado");
                $produtoValido = false;
            }else{
                $prato = $this->consultaPrato($cdfilauxprat, $cdprato);
                if(!$prato){
                    $this->incluiMensagem("Produto $identificacao - Prato informado não cadastrado");
                    $produtoValido = false;
                }
            }
        }
        
        if($this->controlaAlmoxarifado) {
            if($cdalmoxarife == ' '){
                $this->incluiMensagem("Produto $identificacao - Almoxarifado não informado");
                $produtoValido = false;
            } else{
                $almoxarifado = $this->consultaAlmoxarifado($cdfilial, $cdalmoxarife);
                if(!$almoxarifado){
                    $this->incluiMensagem("Produto $identificacao - Almoxarifado informado não cadastrado");
                    $produtoValido = false;
                }
            }
        }
        
        if($this->controlaLocal) {
            if($cdlocalestoq == ' '){
                $this->incluiMensagem("Produto $identificacao - Local não informado");
                $produtoValido = false;
            } else{
                $local = $this->consultaLocalEstoque($cdfilial, $cdalmoxarife, $cdlocalestoq);
                if(!$local){
                    $this->incluiMensagem("Produto $identificacao - Local informado não cadastrado");
                    $produtoValido = false;
                }
            }
        }
        
        if($this->controlaLote) {
            if($nrloteestq == ' '){
                $this->incluiMensagem("Produto $identificacao - Lote não informado");
                $produtoValido = false;
            } else{
                $lote = $this->consultaLote($cdfilial, $dtretirada, $cdproduto, $nrloteestq, $nrsublote);
                if(!$lote){
                    $this->incluiMensagem("Produto $identificacao - Lote informado não cadastrado");
                    $produtoValido = false;
                }
            }
        }
        
        //Se não sobrepõe, incrementa as quantidades já lançadas para o produto
        if(!$sobrepor){
            $qtExtra = round($qtExtraJaLancada + $qtExtra, $this->precisao);
            $qtDevolvida = round($qtDevolvidaJaLancada + $qtDevolvida, $this->precisao);
        }
        else{
            $this->incluiMensagem("Produto $identificacao foi sobreposto");    
        }
        
        $qtRetirada = $qtPrevista - $qtDevolvida + $qtExtra;
        
        $qtsaldoretirada = 0.0;
        if($qtRetirada <= 0){
            $qtRetiradaExibicao = strval($qtRetirada);
            $this->incluiMensagem("Produto $identificacao com quantidade a retirar ( $qtRetiradaExibicao ) menor ou igual a zero");
            $produtoValido = false;
        } else {
            $qtsaldoretirada = $this->consultaSaldoDia($cdfilial, $dtretirada, $cdproduto, $cdalmoxarife, $cdlocalestoq, $nrloteestq, $nrsublote);
            if($idevolestoq == 'S'){
                $qtsaldoretirada = $qtsaldoretirada + $qtRetiradaJaLancada;
            }
            if($qtRetirada > $qtsaldoretirada) {
                $qtRetiradaExibicao = strval($qtRetirada);
                if($this->visualizaSaldo){
                    $saldo = strval($qtsaldoretirada);
                    $this->incluiMensagem("Produto $identificacao com quantidade a retirar ( $qtRetiradaExibicao ) maior que a quantidade disponível em estoque ( $saldo )");
                }
                else{
                    $this->incluiMensagem("Produto $identificacao com quantidade a retirar ( $qtRetiradaExibicao ) maior que a quantidade disponível em estoque");
                }
                $produtoValido = false;
            }
        }
        
        if(!empty($cdproduto)){
            $produtoTeknisa = $this->consultaProduto($produto['CDPRODUTO'], $cdfilial);
            if(!$produtoTeknisa){
                $this->incluiMensagem("Produto $identificacao não cadastrado");
                return false;
            }
            $produtoEstoque = ($produtoTeknisa['CDPRODUTO'] == $produtoTeknisa['CDPRODESTO']);
            $produtoControlaEstoque = ($produtoTeknisa['IDCNTRESTOQ'] == 'S');
            $ativo = ($produtoTeknisa['IDPRODATIV'] == 'A') || ($produtoTeknisa['IDPRODATIVPR'] == 'S');
            if(!$ativo){
                $this->incluiMensagem("Produto $identificacao não está ativo na unidade");
                $produtoValido = false;
            }
            if(!$produtoEstoque){
                $this->incluiMensagem("Produto $identificacao não é um produto de estoque");
                $produtoValido = false;
            }
            if(!$produtoControlaEstoque){
                $this->incluiMensagem("Produto $identificacao não controla estoque para unidade");
                $produtoValido = false;
            }
        }
        
        //Se o tipo de retirada padrão não foi informado, não segue com a inclusão do produto (mas permite a validação dos dados do mesmo)
        if(empty($cdtpretipadr)){
            $produtoValido = false;
        }
        
        if(!$produtoValido) return false;
        
        return [
            'CDPRODUTO' => $cdproduto,
            'CDFILIAL' => $cdfilial,
            'NRCARDAPIO' => $nrcardapio,
            'NRRETIRADA' => $nrretirada,
            'NRLANCESTQ' => $nrlancestq,
            'DTRETIRADA' => $dtretirada,
            'CDBARPRODUTO' => $cdbarproduto,
            'CDFILAUXPRAT' => $cdfilauxprat,
            'CDPRATO' => $cdprato,
            'CDTIPOPRAT' => $cdtipoprat,
            'CDALMOXARIFE' => $cdalmoxarife,
            'CDLOCALESTOQ' => $cdlocalestoq,
            'CDFILANCESTQ' => $cdfilancestq,
            'NRLOTEESTQ' => $nrloteestq,
            'NRSUBLOTE' => $nrsublote,
            'NRSEQUITRETI' => $nrsequitreti, 
            'NRSEQUITEM' => $nrsequitem,
            'QTPREVISRETI' => $qtPrevista,
            'QTPREVISRETI_ANTERIOR' => $qtPrevistaJaLancada,
            'QTEXTRARETI' => $qtExtra,
            'QTEXTRARETI_ANTERIOR' => $qtExtraJaLancada,
            'QTDEVOLVRETI' => $qtDevolvida,
            'QTDEVOLVRETI_ANTERIOR' => $qtDevolvidaJaLancada,
            'QTRETIRADA' => $qtRetirada,
            'QTRETIRADA_ANTERIOR' => $qtRetiradaJaLancada,
            'CDTPRETIPADR' => $cdtpretipadr,
            'QTSALDORETIRADA' => $qtsaldoretirada,
            'QTSALDORETIRADA_ANTERIOR' => $qtSaldoRetiradaJaLancado,
            'QTLANCTOEST_ANTERIOR' => $qtlanctoest,
            'QTTOTLANCTO_ANTERIOR' => $qttotlancto,
            'VRUNILANCTO_ANTERIOR' => $vrunilancto,
            'VRTOTLANCTO_ANTERIOR' => $vrtotlancto,
            'IDEVOLESTOQ' => 'N',
            'IDEVOLESTOQ_ANTERIOR' => $idevolestoq,
            'IDENTIFICACAO' => $identificacao,
            'PRODUTO_NOVO' => $produtoNovo
        ];
    }
    //Popula array de produtos alterados (verifica duplicidades)
    private function registraProdutosAlteradosSemDuplicidade(&$produtos_aplicados, $produto){
        foreach($produtos_aplicados as $prodAlt){
            if(
                ($prodAlt['CDFILIAL'] == $produto['CDFILIAL']) &&
                ($prodAlt['NRCARDAPIO'] == $produto['NRCARDAPIO']) &&
                ($prodAlt['NRRETIRADA'] == $produto['NRRETIRADA']) &&
                ($prodAlt['CDPRODUTO'] == $produto['CDPRODUTO']) &&
                ($prodAlt['NRLANCESTQ'] == $produto['NRLANCESTQ']) &&
                ($prodAlt['CDTIPOPRAT'] == $produto['CDTIPOPRAT']) &&
                ($prodAlt['CDFILAUXPRAT'] == $produto['CDFILAUXPRAT']) &&
                ($prodAlt['CDPRATO'] == $produto['CDPRATO'])
            ){
                $identProd = $produto['IDENTIFICACAO'];
                $identProdAlt = $prodAlt['IDENTIFICACAO'];
                $this->incluiMensagem("Produto $identProd em duplicidade com o produto $identProdAlt");
                return false;
            }
        }
        $produtos_aplicados[] = $produto;
        return true;
    }
    //Insere item de estoque do lançamento
    private function insereItemEstoque($produto){
        $vrunilancto = $this->consultaCustoEstoque($produto['CDFILIAL'], $produto['CDPRODUTO'], $produto['DTRETIRADA'],
                              $produto['CDALMOXARIFE'], $produto['CDLOCALESTOQ'], $produto['NRLOTEESTQ'], $produto['NRSUBLOTE']);
        $vrtotlancto = $produto['QTRETIRADA'] * $vrunilancto;
        $query = "
            INSERT INTO ITLANCTOEST
            (CDFILIAL, NRLANCESTQ, NRSEQUITEM, CDPRODUTO,
             QTLANCTOEST, IDTIPOMOVI, DTLANCMOVI, IDORDEMCLC,
             QTTOTLANCTO, CDFILIMOVI, CDPRODMOVI, IDEVOLESTOQ,
             CDALMOXARIFE, CDLOCALESTOQ, NRLOTEESTQ, NRSUBLOTE,
             CDTIPORETI, VRUNILANCTO, VRTOTLANCTO,
             NRORG, NRORGULTATU, NRORGINCLUSAO,
             DTULTATU, DTINCLUSAO,
             CDOPERULTATU, CDOPERINCLUSAO)
            VALUES 
            (:CDFILIAL, :NRLANCESTQ, :NRSEQUITEM, :CDPRODUTO,
             :QTLANCTOEST, :IDTIPOMOVI, TO_DATE(:DTLANCMOVI, 'YYYYMMDD'), :IDORDEMCLC,
             :QTTOTLANCTO, :CDFILIMOVI, :CDPRODMOVI, :IDEVOLESTOQ,
             :CDALMOXARIFE, :CDLOCALESTOQ, :NRLOTEESTQ, :NRSUBLOTE,
             :CDTIPORETI, :VRUNILANCTO, :VRTOTLANCTO,
             :NRORG, :NRORG, :NRORG,
             SYSDATE, SYSDATE,
             :CDOPERADORLOGADO, :CDOPERADORLOGADO)
        ";
        $params = [
            "CDFILIAL" => $produto['CDFILANCESTQ'],
            "NRLANCESTQ" => $produto['NRLANCESTQ'],
            "NRSEQUITEM" => $produto['NRSEQUITRETI'],
            "CDPRODUTO" => $produto["CDPRODUTO"],
            "QTLANCTOEST" => $produto['QTRETIRADA'],
            "IDTIPOMOVI" => '3',
            "DTLANCMOVI" => $produto['DTRETIRADA'],
            "IDORDEMCLC" => '9',
            "QTTOTLANCTO" => $produto['QTRETIRADA'],
            "CDFILIMOVI" => $produto['CDFILIAL'],
            "CDPRODMOVI" => $produto["CDPRODUTO"],
            "IDEVOLESTOQ" => 'N',
            "CDALMOXARIFE" => $produto["CDALMOXARIFE"],
            "CDLOCALESTOQ" => $produto["CDLOCALESTOQ"],
            "NRLOTEESTQ" => $produto["NRLOTEESTQ"],
            "NRSUBLOTE" => $produto["NRSUBLOTE"],
            "CDTIPORETI" => $produto["CDTPRETIPADR"],
            "VRUNILANCTO" => $vrunilancto,
            "VRTOTLANCTO" => $vrtotlancto,
            "NRORG" => $this->environment->getNrOrg(),
            "CDOPERADORLOGADO" => $this->environment->getCdOperador()
        ];
        $prepared = $this->connection->prepare($query);
        $prepared->execute($params);
    }
    //Atualiza item de estoque do lançamento
    private function atualizaItemEstoque($produto, $restaurando = false){
        $qtlanctoest = $produto['QTLANCTOEST_ANTERIOR'];
        $qttotlancto = $produto['QTTOTLANCTO_ANTERIOR'];
        $vrunilancto = $produto['VRUNILANCTO_ANTERIOR'];
        $vrtotlancto = $produto['VRTOTLANCTO_ANTERIOR'];
        $idevolestoq = $produto['IDEVOLESTOQ_ANTERIOR'];
        if(!$restaurando){
            $qtlanctoest = $produto['QTRETIRADA'];
            $qttotlancto = $produto['QTRETIRADA'];
            $idevolestoq = 'N';
            $vrunilancto = $this->consultaCustoEstoque($produto['CDFILIAL'], $produto['CDPRODUTO'], $produto['DTRETIRADA'],
                                  $produto['CDALMOXARIFE'], $produto['CDLOCALESTOQ'], $produto['NRLOTEESTQ'], $produto['NRSUBLOTE']);
            $vrtotlancto = $produto['QTRETIRADA'] * $vrunilancto;
        }
        $query = "
            UPDATE ITLANCTOEST
               SET QTLANCTOEST = :QTLANCTOEST,
                   QTTOTLANCTO = :QTTOTLANCTO,
                   VRUNILANCTO = :VRUNILANCTO,
                   VRTOTLANCTO = :VRTOTLANCTO,
                   IDEVOLESTOQ = :IDEVOLESTOQ,
                   DTULTATU = SYSDATE,
                   CDOPERULTATU = :CDOPERADORLOGADO
             WHERE CDFILIAL = :CDFILIAL
               AND NRLANCESTQ = :NRLANCESTQ
               AND NRSEQUITEM = :NRSEQUITEM
        ";
        $params = [
            "QTLANCTOEST" => $qtlanctoest,
            "QTTOTLANCTO" => $qttotlancto,
            "VRUNILANCTO" => $vrunilancto,
            "VRTOTLANCTO" => $vrtotlancto,
            "IDEVOLESTOQ" => $idevolestoq,
            "CDFILIAL" => $produto['CDFILANCESTQ'],
            "NRLANCESTQ" => $produto['NRLANCESTQ'],
            "NRSEQUITEM" => $produto['NRSEQUITRETI'],
            "CDOPERADORLOGADO" => $this->environment->getCdOperador()
        ];
        $prepared = $this->connection->prepare($query);
        $prepared->execute($params);       
    }
    //Exclui item de estoque do lançamento
    private function excluiItemEstoque($produto){
        $query = "
            DELETE FROM ITLANCTOEST
             WHERE CDFILIAL = :CDFILIAL
               AND NRLANCESTQ = :NRLANCESTQ
               AND NRSEQUITEM = :NRSEQUITEM
        ";
        $params = [
            "CDFILIAL" => $produto['CDFILANCESTQ'],
            "NRLANCESTQ" => $produto['NRLANCESTQ'],
            "NRSEQUITEM" => $produto['NRSEQUITRETI']
        ];
        $prepared = $this->connection->prepare($query);
        $prepared->execute($params);
    }
    //Insere item da retirada
    private function insereItemReti($produto){
        //Insere o item de estoque
        $this->insereItemEstoque($produto);
        
        $query = "
            INSERT INTO ITEMRETI
            (CDFILIAL, NRRETIRADA, NRCARDAPIO, CDPRODUTO,
             QTPREVISRETI, QTEXTRARETI, QTDEVOLVRETI, QTRETIRADA,
             NRSEQUITRETI, CDFIITLANEST, NRLANCESTQ, NRSEQUITEM,
             CDTIPORETI, CDTIPOPRAT, CDPRATO, CDFILAUXPRAT, QTSALDORETIRADA,
             NRORG, NRORGULTATU, NRORGINCLUSAO,
             DTULTATU, DTINCLUSAO,
             CDOPERULTATU, CDOPERINCLUSAO)
            VALUES 
            (:CDFILIAL, :NRRETIRADA, :NRCARDAPIO, :CDPRODUTO,
             :QTPREVISRETI, :QTEXTRARETI, :QTDEVOLVRETI, :QTRETIRADA,
             :NRSEQUITRETI, :CDFIITLANEST, :NRLANCESTQ, :NRSEQUITEM,
             :CDTIPORETI, :CDTIPOPRAT, :CDPRATO, :CDFILAUXPRAT, :QTSALDORETIRADA,
             :NRORG, :NRORG, :NRORG,
             SYSDATE, SYSDATE,
             :CDOPERADORLOGADO, :CDOPERADORLOGADO)
        ";
        $params = [
            "CDFILIAL" => $produto["CDFILIAL"],
            "NRRETIRADA" => $produto["NRRETIRADA"],
            "NRCARDAPIO" => $produto["NRCARDAPIO"],
            "CDPRODUTO" => $produto["CDPRODUTO"],
            "QTPREVISRETI" => $produto["QTPREVISRETI"],
            "QTEXTRARETI" => $produto['QTEXTRARETI'],
            "QTDEVOLVRETI" => $produto['QTDEVOLVRETI'],
            "QTRETIRADA" => $produto['QTRETIRADA'],
            "NRSEQUITRETI" => $produto['NRSEQUITRETI'],
            "CDFIITLANEST" => $produto["CDFILANCESTQ"],
            "NRLANCESTQ" => $produto["NRLANCESTQ"],
            "NRSEQUITEM" => $produto['NRSEQUITRETI'],
            "CDTIPORETI" => $produto["CDTPRETIPADR"],
            "CDTIPOPRAT" => $produto["CDTIPOPRAT"],
            "CDPRATO" => $produto["CDPRATO"],
            "CDFILAUXPRAT" => $produto["CDFILAUXPRAT"],
            "QTSALDORETIRADA" => $produto["QTSALDORETIRADA"],
            "NRORG" => $this->environment->getNrOrg(),
            "CDOPERADORLOGADO" => $this->environment->getCdOperador()
        ];
        $prepared = $this->connection->prepare($query);
        $prepared->execute($params);    
    }
    //Atualiza item da retirada
    private function atualizaItemReti($produto, $restaurando = false){
        $this->atualizaItemEstoque($produto, $restaurando);
        
        $qtprevisreti = $produto['QTPREVISRETI_ANTERIOR'];
        $qtextrareti = $produto['QTEXTRARETI_ANTERIOR'];
        $qtdevolvreti = $produto['QTDEVOLVRETI_ANTERIOR'];
        $qtretirada = $produto['QTRETIRADA_ANTERIOR'];
        $qtsaldoretirada = $produto['QTSALDORETIRADA_ANTERIOR'];
        $idevolestoq = $produto['IDEVOLESTOQ_ANTERIOR'];
        if(!$restaurando){
            $qtprevisreti = $produto['QTPREVISRETI'];
            $qtextrareti = $produto['QTEXTRARETI'];
            $qtdevolvreti = $produto['QTDEVOLVRETI'];
            $qtretirada = $produto['QTRETIRADA'];
            $qtsaldoretirada = $produto['QTSALDORETIRADA'];
            $idevolestoq = $produto['IDEVOLESTOQ'];
        }
        
        $query = "
            UPDATE ITEMRETI
               SET QTPREVISRETI = :QTPREVISRETI,
                   QTEXTRARETI = :QTEXTRARETI,
                   QTDEVOLVRETI = :QTDEVOLVRETI,
                   QTRETIRADA = :QTRETIRADA,
                   QTSALDORETIRADA = :QTSALDORETIRADA,
                   DTULTATU = SYSDATE,
                   CDOPERULTATU = :CDOPERADORLOGADO
             WHERE CDFILIAL = :CDFILIAL
               AND NRCARDAPIO = :NRCARDAPIO
               AND NRRETIRADA = :NRRETIRADA
               AND NRSEQUITRETI = :NRSEQUITRETI
        ";
        $params = [
            "QTPREVISRETI" => $produto["QTPREVISRETI"],
            "QTEXTRARETI" => $produto['QTEXTRARETI'],
            "QTDEVOLVRETI" => $produto['QTDEVOLVRETI'],
            "QTRETIRADA" => $produto['QTRETIRADA'],
            "QTSALDORETIRADA" => $produto["QTSALDORETIRADA"],
            "CDOPERADORLOGADO" => $this->environment->getCdOperador(),
            "CDFILIAL" => $produto["CDFILIAL"],
            "NRRETIRADA" => $produto["NRRETIRADA"],
            "NRCARDAPIO" => $produto["NRCARDAPIO"],
            "NRSEQUITRETI" => $produto['NRSEQUITRETI']
        ];
        $prepared = $this->connection->prepare($query);
        $prepared->execute($params);    
    }
    //Exclui item da retirada
    private function excluiItemReti($produto){
        $query = "
            DELETE FROM ITEMRETI
             WHERE CDFILIAL = :CDFILIAL
               AND NRCARDAPIO = :NRCARDAPIO
               AND NRRETIRADA = :NRRETIRADA
               AND NRSEQUITRETI = :NRSEQUITRETI
        ";
        $params = [
            "CDFILIAL" => $produto['CDFILIAL'],
            "NRCARDAPIO" => $produto['NRCARDAPIO'],
            "NRRETIRADA" => $produto['NRRETIRADA'],
            "NRSEQUITRETI" => $produto['NRSEQUITRETI']
        ];
        $prepared = $this->connection->prepare($query);
        $prepared->execute($params);
        
        $this->excluiItemEstoque($produto);
    }
    //Desfaz as alterações realizadas no produto, se for produto novo, exclui
    private function desfazAlteracoesProduto($produto){
        if($produto['PRODUTO_NOVO']){
            $this->excluiItemReti($produto);
        } else {
            $this->atualizaItemReti($produto, true);
        }
    }
    //Aplica rotina de estoque, retorna false caso haja irregularidades;
    private function atualizaEstoque($cdfilial, $nrlancestq, &$produtos_alterados, $nrcardapio, $nrretirada, $cdfilancestq){
        $estoqueValido = true;
        try{
            $this->ValidacoesEstoqueService->atualizaEstoque($cdfilancestq, $nrlancestq);    
        }catch(\Exception $e){
            $estoqueValido = false;
            //Se alguma transação ainda estiver ativa e está marcado para rollback, elimina as transações ainda pendentes
            while($this->connection->isTransactionActive() && $this->connection->isRollbackOnly())
                $this->connection->rollback();
        }
        if(!$estoqueValido){
            try{
                //É necessária nova transação visto que a rotina de atualização de estoque também realiza transação própria 
                $this->connection->beginTransaction();
                $produtosIrregulares = $this->consultaItensReti($cdfilial, $nrcardapio, $nrretirada, true);
                foreach($produtosIrregulares as $produto){
                    $produtoSemEstoque = array_filter($produtos_alterados, function($prodAlt) use($produto) {
                        return ($prodAlt['NRSEQUITRETI'] == $produto['NRSEQUITRETI']);    
                    });

                    if(count($produtoSemEstoque) > 0){
                        $produtoSemEstoque = array_pop($produtoSemEstoque);
                        $this->desfazAlteracoesProduto($produtoSemEstoque);
                        $identificacao = $produtoSemEstoque['IDENTIFICACAO'];
                        $this->incluiMensagem("Produto $identificacao ficaria irregular no estoque");
                    }
                }
                $produtosRetirada = $this->consultaItensReti($cdfilial, $nrcardapio, $nrretirada);
                if(count($produtosRetirada) == 0){
                    $this->excluiRetirada($cdfilial, $nrcardapio, $nrretirada);
                    $this->incluiMensagem("Retirada ficaria sem produtos válidos e por isso foi excluída", true);
                }
                
                $this->connection->commit();
            }catch(\Exception $e){
                $this->connection->rollback();
                throw $e;
            }
        }
    }
    //Realiza a retirada do serviço e produtos
    private function realizaRetirada($cdfilial, $cdservfili, $dtretirada, $dslancestq, $cdoperador, $produtos){
        try{
            $this->connection->beginTransaction();
            $cdtpretipadr = null;
            $paramplanejaValido = $this->validaParamPlaneja($cdfilial, $cdtpretipadr);
            if(!$paramplanejaValido) return false;
            
            $retirada = $this->consultaOuInsereRetirada($cdfilial, $cdservfili, $dtretirada, $cdoperador, $dslancestq);
            $nrcardapio = $retirada['NRCARDAPIO'];
            $nrretirada = $retirada['NRRETIRADA'];
            $cdfilancestq = $retirada['CDFILANCESTQ'];
            $nrlancestq = $retirada['NRLANCESTQ'];
            $seq = $this->consultaProxNrItem($cdfilial, $nrcardapio, $nrretirada);
            
            /*Produtos que foram alterados/incluidos são armazenados aqui. 
            Ao evoluir o estoque, se algum deles apresentar erro, apaga/restaura linha dos mesmos*/
            $produtos_aplicados = array();
            $i = 0;
            $numProdutos = count($produtos);
            $semDuplicidade = true;
            /*Percorre cada um dos produtos enviados, e se forem válidos segue para a atualização do banco de dados*/
            for($i = 0; $i < $numProdutos; $i++){
                $produto = $this->validaERetornaProduto($produtos[$i], $i, $cdfilial, $nrcardapio, 
                                                        $nrretirada, $dtretirada, $nrlancestq, $cdtpretipadr, $cdfilancestq);
                if($produto){
                    if($produto['PRODUTO_NOVO']){
                        $produto['NRSEQUITRETI'] = str_pad(strval($seq), 4, "0", STR_PAD_LEFT);
                        $seq++;
                    }
                    if($this->registraProdutosAlteradosSemDuplicidade($produtos_aplicados, $produto)){
                        //Se novo produto cria novo sequencial a partir da numeração já existente na retirada
                        if($produto['PRODUTO_NOVO']){
                            $this->insereItemReti($produto);
                        } else{
                            $this->atualizaItemReti($produto);
                        }
                    }
                }
            }
            
            //Se nenhum dos produtos informados foi válido para registro na retirada
            if(count($produtos_aplicados) == 0){
                $this->incluiMensagem("Solicitação sem produtos válidos para realizar retirada", true);
            }
            
            //Se chegou até o final com o status de SUCESO, aplica o commit, rollback caso contrário
            if($this->Utils->statusAtual() == 'S'){
                $this->connection->commit();
            }
            else{
                $this->connection->rollback();
            }
        }catch(\Exception $e){
            $this->connection->rollback();
            throw $e;
        }
        
        //Aplica atualização se estoque caso a operação de retirada tenha sido realizada com sucesso
        if($this->Utils->statusAtual() == 'S'){
            //Atualiza o estoque e verifica possíveis irregularidades
            $this->atualizaEstoque($cdfilial, $nrlancestq, $produtos_aplicados, $nrcardapio, $nrretirada, $cdfilancestq);
        }
    }

    /***
     * Endpoint: pla/servico
     * Retorna lista de serviços para a filial indicada
     * Parâmetros obrigatórios
     *      CDFILIAL
     * Retorno: array de serviços
     * ***/
    public function endpointGetServico($request, $response){
        try{
            $cdfilial = $this->Utils->parametro($request, 'CDFILIAL');
            $data = $this->consultaServicos($cdfilial);
            $this->Utils->retornoJSON($response, $data);
        }catch(\Exception $e){
            if($this->Utils->modoDesenvolvedor())
                throw $e;
            $this->Utils->incluiMensagem("Ocorreu um erro não esperado na execução da rotina. Verifique os parâmetros e tente novamente", true);
            $this->Utils->retornoJSONStatusMensagem($response);
        }
    }
    
    /***
     * Endpoint: pla/retirada
     * Realiza retirada da lista de produtos informada
     * Parâmetros obrigatórios:
     *      CDFILIAL, CDSERVFILI, DTRETIRADA,
     *      DSLANCESTQ, CDOPERADOR,
     *      produtos: array de objetos
     *
     * Parâmetros opcionais:
     *      CUSTOMVAR1, CUSTOMVAR2, CUSTOMVAR3,
     *      CUSTOMDATE1, CUSTOMFLOAT1, CUSTOMFLOAT2,
     *      CUSTOMBOOL1, CUSTOMCLOB1
     *      
     * Retorno: Objeto com os valores status (E ou S) e mensagens (array de string) 
     * ***/
    public function endpointSetRetirada($request, $response){
        try{
            $cdfilial = $this->Utils->parametro($request, 'CDFILIAL');
            $cdservfili = $this->Utils->parametro($request, 'CDSERVFILI');
            $dtretirada = $this->Utils->parametro($request, 'DTRETIRADA');
            $dslancestq = $this->Utils->parametro($request, 'DSLANCESTQ');
            $cdoperador = $this->Utils->parametro($request, 'CDOPERADOR');
            $produtos = $this->Utils->parametro($request, 'produtos');
            
            $CUSTOMVAR1 = $this->Utils->parametro($request, 'CUSTOMVAR1');
            $CUSTOMVAR2 = $this->Utils->parametro($request, 'CUSTOMVAR2');
            $CUSTOMVAR3 = $this->Utils->parametro($request, 'CUSTOMVAR3');
            $CUSTOMDATE1 = $this->Utils->parametro($request, 'CUSTOMDATE1');
            $CUSTOMFLOAT1 = $this->Utils->parametro($request, 'CUSTOMFLOAT1');
            $CUSTOMFLOAT2 = $this->Utils->parametro($request, 'CUSTOMFLOAT2');
            $CUSTOMBOOL1 = $this->Utils->parametro($request, 'CUSTOMBOOL1');
            $CUSTOMCLOB1 = $this->Utils->parametro($request, 'CUSTOMCLOB1');
            
            $this->prefixo = "Unidade: $cdfilial Serviço: $cdservfili Data: $dtretirada - ";
            
            $filialValida = $this->validaFilial($cdfilial);
            $operadorValido = $this->validaOperador($cdoperador);
            $descricaoValida = $this->validaDescricaoEstoque($dslancestq);
            if($filialValida && $operadorValido && $descricaoValida){
                $servicoValido = $this->validaServico($cdfilial, $cdservfili);
                
                $paramfilial = $this->consultaParamfilial($cdfilial);
                $this->controlaAlmoxarifado = ($paramfilial['IDUTILALMOX'] == 'S' ) ? true : false;
                $this->controlaLocal = ($paramfilial['IDUTILLCESTQ'] == 'S' ) ? true : false;
                $this->controlaLote = ($paramfilial['IDUTILLOTE'] == 'S') ? true : false;
                $this->visualizaSaldo = ($paramfilial['IDPERMVISPOSEST'] == 'S') ? true : false;
                
                $periodoValido = $this->validaPeriodo($cdfilial, $dtretirada, $paramfilial);
                $produtosValido = $this->validaListaProdutos($produtos);
                if($servicoValido && $periodoValido && $produtosValido){
                    $this->realizaRetirada($cdfilial, $cdservfili, $dtretirada,
                                          $dslancestq, $cdoperador, $produtos);
                }
            }
            
            $this->Utils->retornoJSONStatusMensagem($response);
        }catch(\Exception $e){
            if($this->Utils->modoDesenvolvedor())
                throw $e;
            $this->Utils->incluiMensagem("Ocorreu um erro não esperado na execução da rotina. Verifique os parâmetros e tente novamente", true);
            $this->Utils->retornoJSONStatusMensagem($response);
        }
    }
}